import argparse
import sys
from pathlib import Path

from .utils import command_exists, get_asset_names, get_release_map, run_command


def download_assets(release, assets, output_dir=None, repo=None):
    """Download a list of assets from a release."""
    print(f"Downloading {len(assets)} asset(s) from release '{release}'")
    command = ['gh', 'release', 'download', release]
    for asset in assets:
        command.extend(['-p', asset])
    if output_dir:
        command.extend(['--dir', str(output_dir)])
    run_command(command, repo=repo)


def main(argv):
    """Main function to download files from a release."""
    parser = argparse.ArgumentParser(description='Download files from a GitHub release and its supplementary releases.')
    parser.add_argument('--repo', '-g', help='The GitHub repository in the format \'owner/repo\'. If not provided, it will be inferred from the current directory.')
    parser.add_argument('--release', '-r', required=True, help='The base name of the release.')
    parser.add_argument('--file-list', '-f', type=Path, help='Path to a text file with a list of files to download (one file per line).')
    parser.add_argument('--output-dir', '-d', type=Path, help='The directory to save the downloaded files. If not provided, files are downloaded to the current directory.')
    parser.add_argument('--skip-existing', action='store_true', help='Skip downloading files that already exist in the output directory.')
    parser.add_argument('--batch-size', '-b', type=int, default=1, help='The number of files to download in a single batch. (default: 50)')
    parser.add_argument('files', nargs='*', help='File names to download.')
    args = parser.parse_args(argv)

    if not command_exists("gh"):
        print("Error: gh command-line tool is not installed. Please install it to continue.", file=sys.stderr)
        return 1

    if args.output_dir:
        args.output_dir.mkdir(parents=True, exist_ok=True)

    if not args.file_list and not args.files:
        parser.error("No files to download. Please provide a file list with --file-list or specify file names as arguments.")

    files_to_download = set()
    if args.file_list:
        if args.file_list.exists():
            with open(args.file_list, 'r') as f:
                files_to_download.update(line.strip() for line in f if line.strip())
        else:
            print(f"Error: File list '{args.file_list}' not found.", file=sys.stderr)
            return 1
            
    if args.files:
        files_to_download.update(args.files)

    if not files_to_download:
        print("No files to download.")
        return 0

    print("Getting release list")
    release_map = get_release_map(args.release, repo=args.repo)
    
    if not release_map:
        print("No releases found to process.", file=sys.stderr)
        return 1
        
    releases_to_process = list(release_map.values())

    print(f"Will process releases: {releases_to_process}")
    print(f"Files to download: {sorted(list(files_to_download))}")

    asset_to_release_map = {}
    release_to_assets_map = {}
    for release in releases_to_process:
        print(f"Fetching assets from release: {release}")
        assets = get_asset_names(release, repo=args.repo)
        for asset in assets:
            if asset in files_to_download:
                asset_to_release_map[asset] = release
                release_to_assets_map.setdefault(release, []).append(asset)

    output_dir = args.output_dir or Path('.')

    downloaded_assets = set()
    skipped_assets = set()
    failed_assets = set()

    for asset in asset_to_release_map.keys():
        output_file = output_dir / asset
        if args.skip_existing and output_file.exists():
            skipped_assets.add(asset)

    skipped_count = len(skipped_assets)
    if skipped_count > 0:
        print(f"Skipping {skipped_count} files that already exist: {', '.join(list(skipped_assets)[:5])}{'...' if skipped_count > 5 else ''}")


    found_assets = set(asset_to_release_map.keys())
    not_found_assets = files_to_download - found_assets
    
    for asset in sorted(list(not_found_assets)):
        print(f"Asset '{asset}' not found in any of the releases.", file=sys.stderr)

    to_download_count = len(files_to_download) - len(skipped_assets) - len(not_found_assets)

    for release, assets in release_to_assets_map.items():
        assets_to_download_for_release = []
        for asset in sorted(assets):
            if asset not in skipped_assets:
                assets_to_download_for_release.append(asset)

        if not assets_to_download_for_release:
            continue

        batch_size = args.batch_size
        for i in range(0, len(assets_to_download_for_release), batch_size):
            batch = assets_to_download_for_release[i:i+batch_size]
            try:
                download_assets(release, batch, args.output_dir, repo=args.repo)
                downloaded_assets.update(batch)
                print(f"Downloaded {len(downloaded_assets)} of {to_download_count} files...")
            except Exception as e:
                print(f"Failed to download a batch of {len(batch)} files from {release}: {e}", file=sys.stderr)
                for asset in batch:
                    output_file = output_dir / asset
                    if not output_file.exists():
                        failed_assets.add(asset)

    total_count = len(files_to_download)
    downloaded_count = len(downloaded_assets)
    skipped_count = len(skipped_assets)
    not_found_count = len(not_found_assets)
    failed_count = len(failed_assets)

    print("--- Download Summary ---")
    print(f"Total files requested for download: {total_count}")
    print(f"Successfully downloaded files: {downloaded_count}")
    print(f"Skipped existing files: {skipped_count}")
    print(f"Files not found: {not_found_count}")
    print(f"Files failed to download: {failed_count}")
    
    if failed_count > 0:
        return 1
    
    return 0

def cli():
    return main(sys.argv[1:])



import argparse
import csv
import sys
from pathlib import Path

from .utils import command_exists, get_release_map, run_command

def get_assets(release, ext, repo=None):
    """Get all assets with a given extension for a given release."""
    output = run_command([
        'gh',
        'release',
        'view',
        release,
        '--json',
        'assets',
        '-q',
        f'.assets[] | select(.name | endswith("{ext}")) | "\\(.name),\\(.size),\\(.url)"'
    ], repo=repo)
    return output.strip()


def cli():
    """Main function to generate file lists and upload to release."""
    parser = argparse.ArgumentParser(description='Generate file lists and upload to a GitHub release.')
    parser.add_argument('--repo', '-g', help='The GitHub repository in the format \'owner/repo\'. If not provided, it will be inferred from the current directory.')
    parser.add_argument('--release', '-r', required=True, help='The base name of the release.')
    parser.add_argument('--extension', '-e', action='append', help='File extension to filter assets by.')
    args = parser.parse_args()

    if not command_exists("gh"):
        print("Error: gh command-line tool is not installed. Please install it to continue.", file=sys.stderr)
        return 1

    if not args.extension:
        print("No file extensions provided. Please specify at least one extension using --extension.", file=sys.stderr)
        return 1

    print("Getting file list")
    release_map = get_release_map(args.release, repo=args.repo)
    releases_to_process = list(release_map.values())
    
    if not releases_to_process:
        print("No releases found to process.", file=sys.stderr)
        return 1

    print(f"Will process releases: {releases_to_process}")

    all_assets = []
    for release in releases_to_process:
        print(f"Processing release: {release}")
        for ext in args.extension:
            assets = get_assets(release, ext, repo=args.repo)
            if assets:
                # Filter out empty strings that can result from split('\n')
                all_assets.extend([line.strip() for line in assets.split('\n') if line.strip()])

    if not all_assets:
        print("No assets found to process.")
        return 0
        
    csv_file = Path('listing_files.csv')

    with open(csv_file, 'w') as csvfile:
        csv_writer = csv.writer(csvfile)
        csv_writer.writerow(['name', 'size', 'url'])
        for asset_line in all_assets:
            csv_writer.writerow(asset_line.split(',', 2))

    print(f"Uploading {csv_file} to release {args.release}")
    run_command(['gh', 'release', 'upload', args.release, str(csv_file), '--clobber'], repo=args.repo)

    csv_file.unlink()
            
    return 0

if __name__ == '__main__':
    sys.exit(cli())



import re
import subprocess

def command_exists(cmd):
    return subprocess.run(f"command -v {cmd}", shell=True, capture_output=True, text=True).returncode == 0

def run_command(cmd, repo=None):
    """Run a shell command and return its output."""
    if repo and cmd[0] == 'gh':
        cmd = ['gh', '-R', repo] + cmd[1:]

    try:
        print(f"Running command: {' '.join(cmd)}")
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        return result.stdout.strip()
    except subprocess.CalledProcessError as e:
        print(f"Error running command '{' '.join(cmd)}': {e}")
        print('command output stderr:')
        print(e.stderr)
        print('command output stdout:')
        print(e.stdout)
        raise Exception(f"Command '{' '.join(cmd)}' failed with exit code {e.returncode}")

def get_release_map(tag, repo=None):
    output = run_command(["gh", "release", "list", "--json", "tagName", "-q", ".[].tagName", '-L', '500'], repo=repo)

    all_releases = output.strip().split('\n')
    pattern = re.compile(f"^{re.escape(tag)}(-extra(?P<num>[0-9]+))?$")

    rel_map = {}
    for r in all_releases:
        match = pattern.match(r)
        if match is None:
            continue
        g = match.groupdict()
        num = g.get('num')
        if num is not None:
            num = int(num)
            if num == 0:
                raise Exception("Release cannot have '-extra0' suffix")
        else:
            num = 0

        rel_map[num] = r

    return rel_map

def get_repo_name_from_gh(repo=None):
    if repo:
        return repo
    try:
        result = run_command(
            ["gh", "repo", "view", "--json", "nameWithOwner", "-q", ".nameWithOwner"], repo=repo
        )
        return result.strip()
    except Exception as e:
        raise Exception(f"Error getting repository name: {e}")

def get_asset_names(release, repo=None):
    """Get all asset names for a given release."""
    output = run_command([
        'gh',
        'release',
        'view',
        release,
        '--json',
        'assets',
        '-q',
        '.assets[].name'
    ], repo=repo)
    output = output.strip()
    if not output:
        return []

    return output.split('\n')



import argparse
import sys
from pathlib import Path

from .utils import command_exists, get_release_map, get_asset_names, get_repo_name_from_gh, run_command

class CliError(Exception):
    pass

FIRST_RELEASE_MAX_ASSETS = 988
OTHER_RELEASE_MAX_ASSETS = 998

class ReleaseMapper:
    def __init__(self, main_tag):
        self.release_to_assets = {}
        self.main_tag = main_tag
        self.assets_to_releases = {}

    def add_release(self, release):
        if release not in self.release_to_assets:
            self.release_to_assets[release] = set()

    def add_asset(self, asset, release):
        if release not in self.release_to_assets:
            self.release_to_assets[release] = set()

        self.release_to_assets[release].add(asset)

        self.assets_to_releases[asset] = release

    def get_available_releases(self):
        available_releases = []

        for release, assets in self.release_to_assets.items():
            max_assets = FIRST_RELEASE_MAX_ASSETS if release == self.main_tag else OTHER_RELEASE_MAX_ASSETS

            count = len(assets)
            if count < max_assets:
                available_releases.append(release)

        return available_releases

    def get_release_for_asset(self, asset):
        if asset in self.assets_to_releases:
            return self.assets_to_releases[asset]
        return None

def get_next_num(release_map):
    nums = sorted(release_map.keys())

    for n in nums:
        if n + 1 not in nums:
            return n + 1

    return nums[-1] + 1



def get_release_title(tag, repo=None):
    try:
        cmd = ["gh", "release", "view", tag, "--json", "name", "-q", ".name"]
        result = run_command(cmd, repo=repo)
        return result.strip()
    except Exception as e:
        raise CliError(f"Error getting release title for tag {tag}: {e}")

def create_release(next_num, main_tag, repo=None):
    new_release = f"{main_tag}-extra{next_num}"
    print(f"Creating new release '{new_release}'...")
    repo_name = get_repo_name_from_gh(repo)
    main_release_title = get_release_title(main_tag, repo=repo)
    main_release_url = f"https://github.com/{repo_name}/releases/tag/{main_tag}"

    new_release_title = f"{main_release_title} Supplementary{next_num}"
    listing_files_url = f"https://github.com/{repo_name}/releases/download/{main_tag}/listing_files.csv"

    try:
        cmd = ["gh", "release", "create", new_release, "--target", "main", "--title", new_release_title, "--notes", f"Extension of [{main_tag}]({main_release_url})\n\n List of files and their sizes is at [listing_files.csv]({listing_files_url})"]
        run_command(cmd, repo=repo)
        print(f"Created release: {new_release}")
    except Exception as e:
        raise CliError(f"Error creating release: {e}")

    return new_release

def upload_assets(release, file_paths, clobber=False, repo=None):
    """Upload a list of assets to a release."""
    command = ["gh", "release", "upload", release]
    command.extend([str(p) for p in file_paths])

    if clobber:
        command.append("--clobber")
    try:
        print(f"Uploading {len(file_paths)} asset(s) to release '{release}'...")
        run_command(command, repo=repo)
        print(f"Successfully uploaded {len(file_paths)} assets.")
    except Exception as e:
        print(f"Error uploading batch of {len(file_paths)} assets: {e}", file=sys.stderr)
        raise


def cli():
    try:
        parser = argparse.ArgumentParser(
            description="Upload files to a GitHub release, skipping existing files.",
            formatter_class=argparse.RawTextHelpFormatter,
        )
        parser.add_argument('--repo', '-g', help='The GitHub repository in the format \'owner/repo\'. If not provided, it will be inferred from the current directory.')
        parser.add_argument("--release", '-r', help="The git tag of the release to upload to.")
        parser.add_argument("--folder", '-d', help="The local folder containing the files to upload.", type=Path)
        parser.add_argument("--extension", '-e', action='append', help="The extension of the files to upload.")
        parser.add_argument("--create-extra-releases", "-x", action="store_true", help="Create extra releases if needed.")
        parser.add_argument(
            "--overwrite",
            action="store_true",
            help="Allow overwriting existing assets.",
        )
        parser.add_argument('--batch-size', '-b', type=int, default=1, help='The number of files to upload in a single batch. (default: 50)')
        args = parser.parse_args()

        if not command_exists("gh"):
            raise CliError("Error: gh command-line tool is not installed. Please install it to continue.")

        if not args.folder.is_dir():
            raise CliError(f"Error: Folder '{args.folder}' not found.")

        print(f"Fetching existing assets for releases matching pattern '{args.release}(-extra[0-9]+)?'...")

        release_map = get_release_map(args.release, repo=args.repo)
        releases_to_process = list(release_map.values())

        if not releases_to_process:
            raise CliError(f"Error: No releases found matching pattern '{args.release}(-extra[0-9]+)?'.")

        if args.release not in releases_to_process:
            raise CliError(f"Error: Specified main release '{args.release}' not found among fetched releases.")

        print(f"Found releases: {', '.join(releases_to_process)}")

        release_mapper = ReleaseMapper(args.release)

        for rel in releases_to_process:
            release_mapper.add_release(rel)
            print(f"Fetching assets from release: {rel}")
            assets = get_asset_names(rel, repo=args.repo)
            for asset in assets:
                release_mapper.add_asset(asset, rel)

        print(f"Starting upload process from folder '{args.folder}'...")

        files_to_upload_paths = []
        for ext in args.extension or []:
            files_to_upload_paths += sorted(list(args.folder.glob(f"*{ext}")))

        total_files = len(files_to_upload_paths)
        
        files_to_skip = []
        files_to_overwrite = {}  # release -> list of file_paths
        files_for_new_upload = []

        for file_path in files_to_upload_paths:
            filename = file_path.name
            release = release_mapper.get_release_for_asset(filename)

            if release:
                if args.overwrite:
                    files_to_overwrite.setdefault(release, []).append(file_path)
                else:
                    files_to_skip.append(file_path)
            else:
                files_for_new_upload.append(file_path)
        
        skipped_count = len(files_to_skip)
        upload_count_target = total_files - skipped_count
        if skipped_count > 0:
            skipped_filenames = [f.name for f in files_to_skip]
            print(f"Skipping {skipped_count} files that already exist: {', '.join(skipped_filenames[:5])}{'...' if skipped_count > 5 else ''}")

        overwritten_assets = set()
        failed_uploads = set()
        newly_uploaded_assets = set()

        for release, file_paths in files_to_overwrite.items():
            print(f"Overwriting {len(file_paths)} files in release '{release}'...")
            batch_size = args.batch_size
            for i in range(0, len(file_paths), batch_size):
                batch = file_paths[i:i + batch_size]
                try:
                    upload_assets(release, batch, clobber=True, repo=args.repo)
                    overwritten_assets.update(p.name for p in batch)
                    uploaded_count = len(overwritten_assets) + len(newly_uploaded_assets)
                    print(f"Uploaded {uploaded_count} of {upload_count_target} files...")
                except Exception:
                    failed_uploads.update(p.name for p in batch)
        
        uploads_by_release = {}
        
        print(f"Processing {len(files_for_new_upload)} new files to upload...")
        for i, file_path in enumerate(files_for_new_upload, 1):
            filename = file_path.name
            print(f"[{i}/{len(files_for_new_upload)}] Assigning release for: {filename}")
            available_releases = release_mapper.get_available_releases()

            if not available_releases:
                if args.create_extra_releases:
                    next_num = get_next_num(release_map)
                    new_release = create_release(next_num, args.release, repo=args.repo)
                    release_mapper.add_release(new_release)
                    available_releases.append(new_release)
                    release_map[next_num] = new_release
                else:
                    print(f"Error: All existing releases are full. No space to upload '{filename}'. Skipping.", file=sys.stderr)
                    failed_uploads.add(filename)
                    continue

            upload_target = available_releases[0]
            uploads_by_release.setdefault(upload_target, []).append(file_path)
            release_mapper.add_asset(filename, upload_target)

        for release, file_paths in uploads_by_release.items():
            batch_size = args.batch_size
            for i in range(0, len(file_paths), batch_size):
                batch = file_paths[i:i + batch_size]
                try:
                    upload_assets(release, batch, repo=args.repo)
                    newly_uploaded_assets.update(p.name for p in batch)
                    uploaded_count = len(overwritten_assets) + len(newly_uploaded_assets)
                    print(f"Uploaded {uploaded_count} of {upload_count_target} files...")
                except Exception:
                    failed_uploads.update(p.name for p in batch)


        print("Upload process complete.")
        print()
        print("--- Summary ---")
        print(f"Total files processed: {total_files}")
        print(f"Files newly uploaded:  {len(newly_uploaded_assets)}")
        print(f"Files skipped:         {skipped_count}")
        print(f"Files overwritten:     {len(overwritten_assets)}")
        print(f"Files failed to upload(possibly overestimated): {len(failed_uploads)}")

        if len(failed_uploads) > 0:
            print("Failed files:", ", ".join(sorted(list(failed_uploads))))
            return 1

        return 0
    except CliError as e:
        print(e, file=sys.stderr)
        return 1

if __name__ == "__main__":
    sys.exit(cli())



import argparse
import sys
from pathlib import Path

from .utils import command_exists, get_asset_names, get_release_map, run_command

def delete_asset(release, asset_name, repo=None):
    """Delete an asset from a release."""
    print(f"Deleting asset '{asset_name}' from release '{release}'")
    run_command(['gh', 'release', 'delete-asset', release, asset_name, '-y'], repo=repo)

def cli():
    """Main function to delete files from a release."""
    parser = argparse.ArgumentParser(description='Delete files from a GitHub release and its supplementary releases.')
    parser.add_argument('--repo', '-g', help='The GitHub repository in the format \'owner/repo\'. If not provided, it will be inferred from the current directory.')
    parser.add_argument('--release', '-r', required=True, help='The base name of the release.')
    parser.add_argument('--file-list', '-f', type=Path, help='Path to a text file with a list of files to delete (one file per line).')
    parser.add_argument('files', nargs='*', help='File names to delete.')
    args = parser.parse_args()

    if not command_exists("gh"):
        print("Error: gh command-line tool is not installed. Please install it to continue.", file=sys.stderr)
        return 1

    if not args.file_list and not args.files:
        parser.error("No files to delete. Please provide a file list with --file-list or specify file names as arguments.")

    files_to_delete = set()
    if args.file_list:
        if args.file_list.exists():
            with open(args.file_list, 'r') as f:
                files_to_delete.update(line.strip() for line in f if line.strip())
        else:
            print(f"Error: File list '{args.file_list}' not found.", file=sys.stderr)
            return 1
            
    if args.files:
        files_to_delete.update(args.files)

    if not files_to_delete:
        print("No files to delete.")
        return 0

    print("Getting release list")
    release_map = get_release_map(args.release, repo=args.repo)
    
    if not release_map:
        print("No releases found to process.", file=sys.stderr)
        return 1
        
    releases_to_process = list(release_map.values())

    print(f"Will process releases: {releases_to_process}")
    print(f"Files to delete: {sorted(list(files_to_delete))}")

    deleted_assets = set()
    for release in releases_to_process:
        print(f"Processing release: {release}")
        assets = get_asset_names(release, repo=args.repo)
        
        assets_set = set(assets)
        assets_to_delete_from_release = files_to_delete.intersection(assets_set)

        if not assets_to_delete_from_release:
            print(f"No matching files to delete in release '{release}'.")
            continue

        for asset in sorted(list(assets_to_delete_from_release)):
            if delete_asset(release, asset, repo=args.repo):
                deleted_assets.add(asset)

    total_count = len(files_to_delete)
    deleted_count = len(deleted_assets)
    skipped_count = total_count - deleted_count

    print("--- Deletion Summary ---")
    print(f"Total files requested for deletion: {total_count}")
    print(f"Successfully deleted files: {deleted_count}")
    print(f"Skipped files (not found in any release): {skipped_count}")
    
    return 0

if __name__ == '__main__':
    sys.exit(cli())
